//*****************************************************************************
//  Copyright (c) 2021 Glenfly Tech Co., Ltd.. 
//  All Rights Reserved.
//
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Glenfly Tech Co., Ltd..;
//  the contents of this file may not be disclosed to third parties, copied or
//  duplicated in any form, in whole or in part, without the prior written
//  permission of Glenfly Tech Co., Ltd..
//
//  The copyright of the source code is protected by the copyright laws of the People's
//  Republic of China and the related laws promulgated by the People's Republic of China
//  and the international covenant(s) ratified by the People's Republic of China. 
//*****************************************************************************

#include <linux/module.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/font.h>
#include <linux/mutex.h>
#include <linux/videodev2.h>
#include <linux/kthread.h>
#include <linux/freezer.h>
#include <linux/delay.h>
#include <linux/workqueue.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/sort.h>
#include <asm/io.h>
#ifdef   __arm__
#include <asm/system.h>
#endif
//#include <asm/uaccess.h>

#include <media/videobuf2-vmalloc.h>
#include <media/videobuf2-dma-contig.h>
#include <media/v4l2-device.h>
#include <media/v4l2-ioctl.h>
#include <media/v4l2-ctrls.h>
#include <media/v4l2-fh.h>
#include <media/v4l2-event.h>
#include <media/v4l2-common.h>

#include <linux/platform_device.h>

#include "gf_vip_videobuf.h"

#include "v4l2_vip_def.h"
#include "gf_vip_video.h"
#if VIP_V4L2_VERSION_CODE >= KERNEL_VERSION(5,4,0)
#include <media/videobuf2-v4l2.h>
#endif

static int debug = 1;
module_param(debug, int, 0644);


#define dprintk(level, fmt, arg...)					      \
	do {								      \
		if (debug >= level)					      \
			pr_info("vip-v4l2: %s: " fmt, GF_FUNC_NAME(__func__), ## arg); \
	} while (0)


#define IMAGE_WIDTH  1280
#define IMAGE_HEIGHT 720

#define VIP_CAMERA_MAX_NUMBER               8
static struct h8_video_dev *g_vip_dev[VIP_CAMERA_MAX_NUMBER];

static const char  vendor_name[]={"vip_v4l2"};
extern h8v_ctrl_t           h8v_ctrls[];
static struct h8v_fmt formats[] =
{
    {
        .name   = "ARGB8888",
        #if VIP_V4L2_VERSION_CODE < KERNEL_VERSION(4,1,0)
        .fourcc = V4L2_PIX_FMT_RGB32,
        #else
        .fourcc = V4L2_PIX_FMT_ARGB32,
        #endif
        .depth  = 32,
    },
    {
        .name   = "4:2:2, packed, YUYV",
        .fourcc = V4L2_PIX_FMT_YUYV,
        .depth  = 16,
    },
    {
        .name   = "4:2:0, two planes, NV12",
        .fourcc = V4L2_PIX_FMT_NV12,
        .depth  = 12,
    },

};
//static int nFormats = sizeof(formats)/sizeof(struct h8v_fmt);

static bool dma_fence_check(struct drm_gf_gem_object *obj)
{
    dma_fence_t *excl = NULL;
    struct reservation_object *resv = obj->resv;
    bool signaled = true;

    excl = reservation_object_get_excl_rcu(resv);
    if (excl)
    {
        signaled = dma_fence_is_signaled(excl);
    }

    dma_fence_put(excl);

    return signaled;
}

int is_signaled(struct cam_buf *vip_buf)
{
    int signaled = true;
    struct h8v_buffer *buf = container_of(vip_buf, struct h8v_buffer, vip_buf);
    struct drm_gf_gem_object *obj = (struct drm_gf_gem_object *)vb2_plane_cookie(&buf->vb, 0);
    if (!obj) {
        dprintk(1, "gem obj is NULL!!");
        return signaled;
    }

    signaled = dma_fence_check(obj);
    if (!signaled)
        dprintk(1, "dma fence is 0, gpu_virt_addr %llx", obj->info.gpu_virt_addr);

    return signaled;
}

void* get_cpu_vaddr(struct cam_buf *vip_buf)
{
    int signaled = true;
    void *vbuf;
    struct h8v_buffer *buf = container_of(vip_buf, struct h8v_buffer, vip_buf);
    struct drm_gf_gem_object *obj = (struct drm_gf_gem_object *)vb2_plane_cookie(&buf->vb, 0);
    if (!obj) {
        dprintk(1, "gem obj is NULL!!");
        return NULL;
    }

    vbuf = vb2_plane_vaddr(&buf->vb, 0);

    return vbuf;
}

/* ------------------------------------------------------------------
	Videobuf operations
   ------------------------------------------------------------------*/

#if VIP_V4L2_VERSION_CODE < KERNEL_VERSION(4,4,0)
//(0, 4.4.0)
static int queue_setup(struct vb2_queue *vq, const struct v4l2_format *fmt, unsigned int *nbuffers,
				       unsigned int *nplanes, unsigned int sizes[], void *alloc_ctxs[])
{
	struct h8_video_dev *dev = vb2_get_drv_priv(vq);
	unsigned long size;

    dprintk(1, "queue_setup buffer num %u, width %u height %u\n", *nbuffers, dev->width, dev->height);

    //size = dev->width * dev->height * 2;

	if (*nbuffers < 2)
		*nbuffers = 2;

	*nplanes = 1;
    //the image size
    //sizes[0] = IMAGE_WIDTH * IMAGE_HEIGHT * 2;
    sizes[0] = dev->width * dev->height * 2;

	/*
	 * videobuf2-vmalloc allocator is context-less so no need to set
	 * alloc_ctxs array.
	 */

	return 0;
}
#elif VIP_V4L2_VERSION_CODE < KERNEL_VERSION(4,5,7)
//start with 4.4.194, [4.4.0 , 4.5.7)
static int queue_setup(struct vb2_queue *q, const void *parg,
                           unsigned int *num_buffers, unsigned int *num_planes,
                           unsigned int sizes[], void *alloc_ctxs[])
{
	struct h8_video_dev *dev = vb2_get_drv_priv(q);
	unsigned long size;

    dprintk(1, "queue_setup buffer num %u, width %u height %u\n", *num_buffers, dev->width, dev->height);

    //size = dev->width * dev->height * 2;

	if (*num_buffers < 2)
		*num_buffers = 2;

	*num_planes = 1;
    //the image size
    //sizes[0] = IMAGE_WIDTH * IMAGE_HEIGHT * 2;
    sizes[0] = dev->width * dev->height * 2;

	/*
	 * videobuf2-vmalloc allocator is context-less so no need to set
	 * alloc_ctxs array.
	 */

	return 0;
}
#elif VIP_V4L2_VERSION_CODE < KERNEL_VERSION(4,8,0)
//[4.5.7 , 4.8.0)
static int queue_setup(struct vb2_queue *q,
                           unsigned int *num_buffers, unsigned int *num_planes,
                           unsigned int sizes[], void *alloc_ctxs[])
{
	struct h8_video_dev *dev = vb2_get_drv_priv(q);
	unsigned long size;

    dprintk(1, "queue_setup buffer num %u, width %u height %u\n", *num_buffers, dev->width, dev->height);

    //size = dev->width * dev->height * 2;

	if (*num_buffers < 2)
		*num_buffers = 2;

	*num_planes = 1;
    //the image size
    //sizes[0] = IMAGE_WIDTH * IMAGE_HEIGHT * 2;
    sizes[0] = dev->width * dev->height * 2;

	/*
	 * videobuf2-vmalloc allocator is context-less so no need to set
	 * alloc_ctxs array.
	 */

	return 0;
}
#else
//[4.8.0, ~)
static int queue_setup(struct vb2_queue *q,
			   unsigned int *num_buffers, unsigned int *num_planes,
			   unsigned int sizes[], struct device *alloc_devs[])
{
	struct h8_video_dev *dev = vb2_get_drv_priv(q);
	unsigned long size;

    dprintk(1, "queue_setup buffer num %u, width %u height %u\n", *num_buffers, dev->width, dev->height);

    //size = dev->width * dev->height * 2;

	if (*num_buffers < 2)
		*num_buffers = 2;

	*num_planes = 1;
    //the image size
    //sizes[0] = IMAGE_WIDTH * IMAGE_HEIGHT * 2;
    sizes[0] = dev->width * dev->height * 2;

	/*
	 * videobuf2-vmalloc allocator is context-less so no need to set
	 * alloc_ctxs array.
	 */

	return 0;
}
#endif

static int buffer_init(struct vb2_buffer *vb)
{
	struct h8_video_dev *dev   = vb2_get_drv_priv(vb->vb2_queue);
    #if VIP_V4L2_VERSION_CODE >= KERNEL_VERSION(5,4,0)
    struct vb2_v4l2_buffer *vbuf = NULL;
    #endif

    dprintk(2, "fmt name %s\n", dev->fmt->name);
    BUG_ON(NULL == dev->fmt);

	/*
	 * This callback is called once per buffer, after its allocation.
	 *
	 * Vivi does not allow changing format during streaming, but it is
	 * possible to do so when streaming is paused (i.e. in streamoff state).
	 * Buffers however are not freed when going into streamoff and so
	 * buffer size verification has to be done in buffer_prepare, on each
	 * qbuf.
	 * It would be best to move verification code here to buf_init and
	 * s_fmt though.
	 */
    #if VIP_V4L2_VERSION_CODE >= KERNEL_VERSION(5,4,0)
	vbuf = to_vb2_v4l2_buffer(vb);
    //dprintk(1, "buffer_init, fd %d", vbuf->request_fd);
	vbuf->request_fd = -1;  //don't use media controll
    #endif

	return 0;
}

static int buffer_prepare(struct vb2_buffer *vb)
{
	struct h8_video_dev *dev = vb2_get_drv_priv(vb->vb2_queue);
	struct h8v_buffer *buf = container_of(vb, struct h8v_buffer, vb);

	BUG_ON(NULL == dev->fmt);

    //printk("kmd: buffer_prepare\n");
    //printk("kmd: buffer_prepare, size = %lu \n", vb2_plane_size(vb, 0));
	/*
	 * Theses properties only change when queue is idle, see s_fmt.
	 * The below checks should not be performed here, on each
	 * buffer_prepare (i.e. on each qbuf). Most of the code in this function
	 * should thus be moved to buffer_init and s_fmt.
	 */
	if (dev->width  < 48 || dev->width  > MAX_WIDTH ||
	    dev->height < 32 || dev->height > MAX_HEIGHT)
		return -EINVAL;

	buf->fmt = dev->fmt;
	return 0;
}

#if VIP_V4L2_VERSION_CODE < KERNEL_VERSION(4,1,0)
static int buffer_finish(struct vb2_buffer *vb)
{
    //printk("kmd: buffer_finish \n");
	//struct h8_video_dev *dev = vb2_get_drv_priv(vb->vb2_queue);
	return 0;
}
#else
static void buffer_finish(struct vb2_buffer *vb)
{
    #if VIP_V4L2_VERSION_CODE >= KERNEL_VERSION(5,4,0)
	struct vb2_v4l2_buffer *vbuf = NULL;
	vbuf = to_vb2_v4l2_buffer(vb);
	vbuf->request_fd = -1;  //don't use media_device 
    #endif

    return;
}
#endif

static void buffer_cleanup(struct vb2_buffer *vb)
{
    dprintk(3, "no-op\n");
	//struct h8_video_dev *dev = vb2_get_drv_priv(vb->vb2_queue);
}

static struct buf_ops cam_buf_ops = {
    .is_signaled		= is_signaled,
    .get_addr           = NULL,
    .get_cpu_vaddr      = get_cpu_vaddr,
};

//set buffer to vip hw
static void buffer_queue(struct vb2_buffer *vb)
{
	struct h8_video_dev *dev = vb2_get_drv_priv(vb->vb2_queue);
	struct h8v_buffer   *buf = container_of(vb, struct h8v_buffer, vb);
    void *vbuf;
    struct drm_gf_gem_object *obj = NULL;

    //printk("kmd: buffer_queue\n");

    if (dev->stream_stoped)
    {
        dprintk(1, "queue buffer before stream start\n");
    }

    spin_lock(&dev->slock);

    vbuf = vb2_plane_vaddr(&buf->vb, 0);
    obj = (struct drm_gf_gem_object *)vb2_plane_cookie(&buf->vb, 0);
    if (!obj) {
        dprintk(1, "gem obj is NULL!!");
    }

    cam_buf_init(&buf->vip_buf, &cam_buf_ops, obj->info.gpu_virt_addr);

    mgr_queue_buffer(dev->mgr, &buf->vip_buf);

    spin_unlock(&dev->slock);

    dprintk(3, "allocation %x", obj->info.allocation);
}

int buffer_done_callback(void *data, struct cam_buf *buffer, int flag)
{
    struct h8v_buffer *buf = container_of(buffer, struct h8v_buffer, vip_buf);
    if(flag == CAM_BUF_FLAG_RESOLUTION_CHANGED){
        dprintk(1, "receive resolution changed");
#if VIP_V4L2_VERSION_CODE < KERNEL_VERSION(4,4,0)
        buf->vb.v4l2_planes[0].m.fd = -1;
#else
        buf->vb.planes[0].m.fd = -1;
#endif
    }
    vb2_buffer_done(&buf->vb, VB2_BUF_STATE_DONE);

    return 0;
}

static int start_streaming(struct vb2_queue *vq, unsigned int count)
{
	struct h8_video_dev *dev = vb2_get_drv_priv(vq);

    dprintk(1, "start streaming %d\n", count);

    spin_lock(&dev->slock);
    dev->stream_stoped = false;
    spin_unlock(&dev->slock);

    mgr_enable(dev->mgr);
    return 0;
}

/* abort streaming and wait for last buffer */
#if VIP_V4L2_VERSION_CODE < KERNEL_VERSION(4,1,0)
static int stop_streaming(struct vb2_queue *q)
{
    struct h8_video_dev *dev = vb2_get_drv_priv(q);

    spin_lock(&dev->slock);
    if (dev->stream_stoped) {
        dprintk(1, "stream is already stopped \n");
        spin_unlock(&dev->slock);
        return 0;
    }
    dprintk(1, "stop streaming \n");

    dev->stream_stoped = true;
    spin_unlock(&dev->slock);

    mgr_disable(dev->mgr);

	return 0;
}
#else
static void stop_streaming(struct vb2_queue *q)
{
    struct h8_video_dev *dev = vb2_get_drv_priv(q);

    spin_lock(&dev->slock);
    if (dev->stream_stoped) {
        dprintk(1, "stream is already stopped \n");
        spin_unlock(&dev->slock);
        return;
    }
    dprintk(1, "stop streaming \n");

    dev->stream_stoped = true;
    spin_unlock(&dev->slock);

    mgr_disable(dev->mgr);

	return;
}
#endif

static void vip_lock(struct vb2_queue *vq)
{
	struct h8_video_dev *dev = vb2_get_drv_priv(vq);
    //printk("kmd: h8v_lock\n");

	mutex_lock(&dev->mutex);
}

static void vip_unlock(struct vb2_queue *vq)
{
	struct h8_video_dev *dev = vb2_get_drv_priv(vq);
    //printk("kmd: h8v_unlock\n");

	mutex_unlock(&dev->mutex);
}

static struct vb2_ops vip_video_qops = {
	.queue_setup		= queue_setup,
    .wait_prepare       = vip_unlock,
	.wait_finish        = vip_lock,
	.buf_init           = buffer_init,
	.buf_prepare        = buffer_prepare,
	.buf_finish         = buffer_finish,
	.buf_cleanup        = buffer_cleanup,
	.start_streaming    = start_streaming,
	.stop_streaming     = stop_streaming,
    .buf_queue          = buffer_queue,
};


/* ------------------------------------------------------------------
	FILE IO handling
   ------------------------------------------------------------------*/

int cam_mgr_caps_cmp(const void * left, const void * right){
#define EQUAL_FMT(l, r)     (l->format == r->format)
#define EQUAL_SIZE(l, r)    (l->width == r->width && l->height == r->height)
#define EQUAL_FPS(l, r)     (l->fps_num == r->fps_num && l->fps_den == r->fps_den)
#define CMP_FMT(l, r)       (l->format < r->format)
#define CMP_SIZE(l, r)      (((l->width & 0xFFFF) << 16 | (l->height & 0xFFFF)) < ((r->width & 0xFFFF) << 16 | (r->height & 0xFFFF)))
#define CMP_FPS(l, r)       ((l->fps_num ) * (r->fps_den ) < (l->fps_den) * (r->fps_num))

    struct cam_fmt * l = (struct cam_fmt *)left;
    struct cam_fmt * r = (struct cam_fmt *)right;

    if(!EQUAL_FMT(l, r))
        return CMP_FMT(l, r) ? 1 : -1;

    if(!EQUAL_SIZE(l, r))
        return CMP_SIZE(l, r) ? 1 : -1;

    if(!EQUAL_FPS(l, r))
        return CMP_FPS(l, r) ? 1 : -1;

    return 0;
}

void cam_mgr_caps_swap(void * left, void * right, int size){
    struct cam_fmt * l = (struct cam_fmt *)left;
    struct cam_fmt * r = (struct cam_fmt *)right;
    struct cam_fmt tmp;

//    assert(size == sizeof(struct cam_fmt));
    tmp = *l;
    *l = *r;
    *r = tmp;
}

static int vip_video_open(struct file *file)
{
    struct h8_video_dev *dev = video_drvdata(file);
    int    ret;
    int    index = dev->video_index;

    if (dev->opened)
    {
        dprintk(1, "Can't open video%d dev twice.\n", index);
        return -EBUSY;
    }

    ret = v4l2_fh_open(file);
    if (!ret)
        dev->opened = 1;
    else
        dprintk(1, "V4L2 open failed.\n");

    dev->stream_stoped = 1;

    mgr_init(0, index, &dev->mgr);
    mgr_set_callback(dev->mgr, buffer_done_callback, NULL);

    //get format num
    //TODO how to handle if caps_num == 0 ?
    dev->caps.caps_num = mgr_get_caps_num(dev->mgr);
    //malloc memory
    dev->caps.formats = kzalloc(dev->caps.caps_num * sizeof(struct cam_fmt), GFP_KERNEL);
    //get formats
    mgr_get_caps(dev->mgr, &dev->caps);
    //sort formats
    sort(dev->caps.formats, dev->caps.caps_num, sizeof(struct cam_fmt), &cam_mgr_caps_cmp, &cam_mgr_caps_swap);
//    for(i = 0; i < dev->caps.caps_num; i++){
//        tmp = dev->caps.formats + i;
//        dprintk(1, "fmt %x size %dx%d fps %d/%d\n", tmp->format, tmp->width, tmp->height, tmp->fps_num, tmp->fps_den);
//    }
    //default fps init
    dev->fmt_fps.denominator = 0;
    dev->fmt_fps.numerator = 0;
    dprintk(1, "V4L2 open video %d OK.\n", index);
    return ret;
};

static int vip_video_release(struct file *file)
{
    struct h8_video_dev *dev = video_drvdata(file);

    dprintk(1, "release video \n");

    if (v4l2_fh_is_singular_file(file))
        vb2_queue_release(&dev->vb_vidq); //if stream do not stop, this will do stop stream

    mgr_deinit(dev->mgr);

    kfree(dev->caps.formats);
    dev->caps.caps_num = 0;
    dev->opened = 0;

    return v4l2_fh_release(file);
}

static unsigned int vip_video_poll(struct file *file, struct poll_table_struct *wait)
{
    struct h8_video_dev *dev = video_drvdata(file);
    struct v4l2_fh *fh = file->private_data;
    struct vb2_queue *q = &dev->vb_vidq;
    unsigned int res;

    //printk("kmd: h8_video_poll\n");

    res = vb2_poll(q, file, wait);
    if (v4l2_event_pending(fh))
        res |= POLLPRI;
    else
        poll_wait(file, &fh->wait, wait);
    return res;
}

static int vip_file_mmap(struct file *file, struct vm_area_struct *vma)
{
	struct h8_video_dev *dev = video_drvdata(file);
    struct vb2_queue *q = &dev->vb_vidq;

    dprintk(3, "vip_file_mmap\n");
	return vb2_mmap(q, vma);
}

static struct h8v_fmt *get_format(struct v4l2_format *f)
{
    struct h8v_fmt *fmt;
    unsigned int k;

    for (k = 0; k < ARRAY_SIZE(formats); k++)
    {
        fmt = &formats[k];
        if (fmt->fourcc == f->fmt.pix.pixelformat)
            break;
    }

    if (k == ARRAY_SIZE(formats))
        return NULL;

    return &formats[k];
}

/* ------------------------------------------------------------------
	IOCTL vidioc handling
   ------------------------------------------------------------------*/
static int vidioc_querycap(struct file *file, void *priv, struct v4l2_capability *cap)
{
	struct h8_video_dev *dev = video_drvdata(file);

    //V4L2_H8_LOG(video_nr, h8_nr_VIDIOC_QUERYCAP);

    strlcpy(cap->driver, vendor_name, sizeof(cap->driver));
    strlcpy(cap->card, vendor_name, sizeof(cap->card));
    strlcpy(cap->bus_info, dev->v4l2_dev.name, sizeof(cap->bus_info));
    cap->version        = VIP_V4L2_VERSION_CODE;

    cap->device_caps = V4L2_CAP_VIDEO_CAPTURE | V4L2_CAP_STREAMING | V4L2_CAP_READWRITE;
    cap->capabilities = cap->device_caps | V4L2_CAP_DEVICE_CAPS;

	return 0;
}

static int vidioc_enum_fmt_vid_cap(struct file *file, void  *priv, struct v4l2_fmtdesc *f)
{
    struct h8_video_dev *dev = video_drvdata(file);
    struct cam_fmt * tmp;
    uint32_t fmt = 0;
    int index = -1;
    int i;
    if (unlikely(f->index >= dev->caps.caps_num))
        return -EINVAL;
    if(f->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
        return -EINVAL;
    dprintk(2, "vidioc_enum_fmt_vid_cap index %d\n", f->index);

    //strlcpy(f->description, "AA", sizeof(f->description));
    //f->pixelformat = formats[f->index].fourcc;
    for(i = 0; i < dev->caps.caps_num; i++){
        tmp = dev->caps.formats + i;
        if(tmp->format != fmt){
            fmt = tmp->format;
            index++;
            if(index == f->index){
                f->pixelformat = fmt;
                return 0;
            }
        }
    }

    return -EINVAL;
}

static int vidioc_enum_framesizes(struct file *file, void *fh, struct v4l2_frmsizeenum *fsize)
{
    struct h8_video_dev *dev = video_drvdata(file);
    struct cam_fmt * tmp;
    uint32_t fmt = fsize->pixel_format;
    uint32_t width = 0, height = 0;
    int index = -1;
    int i;

    if (unlikely(fsize->index >= dev->caps.caps_num))
        return -EINVAL;

    dprintk(2, "vidioc_enum_framesizes index %d\n", fsize->index);
    for(i = 0; i < dev->caps.caps_num; i++){
        tmp = dev->caps.formats + i;
        if(tmp->format == fmt){
            if(tmp->width != width && tmp->height != height){
                index++;
                width = tmp->width;
                height = tmp->height;
            }
            if(index == fsize->index){
                fsize->discrete.width = width;
                fsize->discrete.height = height;
                fsize->type = V4L2_FRMSIZE_TYPE_DISCRETE;
                return 0;
            }
        }else if(index == -1){
            continue;
        }else{
            return -EINVAL;
        }
    }
    return -EINVAL;
}

static int vidioc_enum_frameintervals(struct file *file, void *fh, struct v4l2_frmivalenum *fival)
{
    struct h8_video_dev *dev = video_drvdata(file);
    struct cam_fmt * tmp;
    uint32_t fmt = fival->pixel_format;
    uint32_t width = fival->width;
    uint32_t height = fival->height;
    int index = -1;
    int i;

    if (unlikely(fival->index >= dev->caps.caps_num))
        return -EINVAL;

    dprintk(2, "vidioc_enum_frameintervals index %d\n", fival->index);
    for(i = 0; i < dev->caps.caps_num; i++){
        tmp = dev->caps.formats + i;
        if(tmp->format == fmt && tmp->width == width && tmp->height == height){
            index++;
            if(index == fival->index){
                // NOTICE! swap num & den
                fival->discrete.numerator = tmp->fps_den;
                fival->discrete.denominator = tmp->fps_num;
                fival->type = V4L2_FRMIVAL_TYPE_DISCRETE;
                return 0;
            }
        }else if(index == -1){
            continue;
        }else{
            return -EINVAL;
        }
    }

    return -EINVAL;
}

static int vidioc_g_fmt_vid_cap(struct file *file, void *priv, struct v4l2_format *f)
{
    struct h8_video_dev *dev = video_drvdata(file);

    f->fmt.pix.width        = dev->width;
    f->fmt.pix.height       = dev->height;
    f->fmt.pix.field        = dev->field;
    f->fmt.pix.pixelformat  = dev->fmt->fourcc;
    f->fmt.pix.bytesperline = 0;//(f->fmt.pix.width * dev->fmt->depth) >> 3;
    f->fmt.pix.sizeimage    = 0;//f->fmt.pix.height * f->fmt.pix.bytesperline;

    if (dev->fmt->fourcc == V4L2_PIX_FMT_YUYV ||
        dev->fmt->fourcc == V4L2_PIX_FMT_UYVY)
        f->fmt.pix.colorspace = V4L2_COLORSPACE_REC709; // V4L2_COLORSPACE_SMPTE170M=BT601
    else
        f->fmt.pix.colorspace = V4L2_COLORSPACE_SRGB;

    dprintk(1, "get format WxH, %dX%d, format %x\n", f->fmt.pix.width, f->fmt.pix.height, f->fmt.pix.pixelformat);
    return 0;
}

static int vidioc_try_fmt_vid_cap(struct file *file, void *priv, struct v4l2_format *f)
{
    //struct h8_video_dev *dev = video_drvdata(file);
    struct h8v_fmt *fmt;
    //enum v4l2_field field;

    //printk("kmd: vidioc_try_fmt_vid_cap\n");

    fmt = get_format(f);
    if (!fmt)
    {
        dprintk(1, "Try_fmt: Fourcc format (0x%08x) invalid.\n", f->type);
        return -EINVAL;
    }

    f->fmt.pix.field = V4L2_FIELD_NONE;//field;
    v4l_bound_align_image(&f->fmt.pix.width, 48, MAX_WIDTH, 2,
                          &f->fmt.pix.height, 32, MAX_HEIGHT, 0, 0);
    f->fmt.pix.bytesperline = (f->fmt.pix.width * fmt->depth) >> 3;
    f->fmt.pix.sizeimage = f->fmt.pix.height * f->fmt.pix.bytesperline;

    if (fmt->fourcc == V4L2_PIX_FMT_YUYV ||
        fmt->fourcc == V4L2_PIX_FMT_NV12)
        f->fmt.pix.colorspace = V4L2_COLORSPACE_REC709;
    else
        f->fmt.pix.colorspace = V4L2_COLORSPACE_SRGB;

    return 0;
}

static int vidioc_s_fmt_vid_cap(struct file *file, void *priv, struct v4l2_format *f)
{
    struct h8_video_dev *dev = video_drvdata(file);
    struct vb2_queue *    q = &dev->vb_vidq;
    struct cam_fmt * tmp;
    int ret, i, best = -1;

    if (vb2_is_streaming(q)) {
        dprintk(1, "vidioc_s_fmt_vid_cap, stream on, can't change\n");
        return -EBUSY;
    }

    for (i = 0; i < dev->caps.caps_num; i++) {
        tmp = dev->caps.formats + i;
        if (f->fmt.pix.pixelformat == tmp->format && f->fmt.pix.width == tmp->width && f->fmt.pix.height == tmp->height) {
            dev->fmt->fourcc = f->fmt.pix.pixelformat;
            dev->width  = f->fmt.pix.width;
            dev->height = f->fmt.pix.height;
            dev->field  = V4L2_FIELD_NONE;//f->fmt.pix.field;

            if(dev->fmt_fps.numerator != 0 && dev->fmt_fps.denominator != 0
                    && dev->fmt_fps.numerator == tmp->fps_num && dev->fmt_fps.denominator == tmp->fps_den){
                best = i;
                break;
            }else if(best == -1)
                best = i;
        }
    }

    if(best != -1){
        dprintk(1, "set format width=%d, height=%d, format=%x, fps=%d/%d\n", dev->width, dev->height,
                f->fmt.pix.pixelformat, dev->caps.formats[best].fps_num, dev->caps.formats[best].fps_den);

        return mgr_set_format(dev->mgr, &dev->caps.formats[best]);
    }

    return -EINVAL;
}


static int vidioc_s_parm(struct file *file, void *fh, struct v4l2_streamparm *a)
{
	struct h8_video_dev *dev = video_drvdata(file);
    int i;

	if (a->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		return -EINVAL;

    if (!a->parm.capture.timeperframe.denominator)
        return -EINVAL;


    for (i = 0; i < dev->caps.caps_num; i++) {
        __u32 format = dev->caps.formats[i].format;
        __u32 width = dev->caps.formats[i].width;
        __u32 height = dev->caps.formats[i].height;

        if (format == dev->fmt->fourcc && width == dev->width && height == dev->height) {
            if (a->parm.capture.timeperframe.numerator * dev->caps.formats[i].fps_num
                    == a->parm.capture.timeperframe.denominator * dev->caps.formats[i].fps_den) {
                dprintk(2, "set fps %d/%d\n", a->parm.capture.timeperframe.denominator, a->parm.capture.timeperframe.numerator);
                dev->fmt_fps.numerator = dev->caps.formats[i].fps_num;
                dev->fmt_fps.denominator = dev->caps.formats[i].fps_den;
                mgr_set_format(dev->mgr, &dev->caps.formats[i]);
                return 0;
            } else {
//                dprintk(1, "set fps failed %d/%d\n", a->parm.capture.timeperframe.denominator, a->parm.capture.timeperframe.numerator);
//                return -EINVAL;
            }

        }
    }

    return -EINVAL;
}

static int vidioc_reqbufs(struct file *file, void *priv, struct v4l2_requestbuffers *p)
{
    struct h8_video_dev *dev = video_drvdata(file);
    int ret;

    mutex_lock(&dev->queue_lock);
    ret = vb2_reqbufs(&dev->vb_vidq, p);
    mutex_unlock(&dev->queue_lock);

    return ret;
}

static int vidioc_querybuf(struct file *file, void *priv, struct v4l2_buffer *p)
{
    struct h8_video_dev *dev = video_drvdata(file);
    int ret;

    mutex_lock(&dev->queue_lock);
    ret = vb2_querybuf(&dev->vb_vidq, p);
    mutex_unlock(&dev->queue_lock);

    return ret;
}

static int vidioc_qbuf(struct file *file, void *priv, struct v4l2_buffer *p)
{
    struct h8_video_dev *dev = video_drvdata(file);
    int ret;

    mutex_lock(&dev->queue_lock);
#if VIP_V4L2_VERSION_CODE < KERNEL_VERSION(4,20,0)
    ret = vb2_qbuf(&dev->vb_vidq, p);
#else
    ret = vb2_qbuf(&dev->vb_vidq, NULL, p);
#endif
    mutex_unlock(&dev->queue_lock);

    return ret;
}

static int vidioc_dqbuf(struct file *file, void *priv, struct v4l2_buffer *p)
{
    struct h8_video_dev *dev = video_drvdata(file);
    int ret;

    mutex_lock(&dev->queue_lock);
    ret = vb2_dqbuf(&dev->vb_vidq, p, file->f_flags & O_NONBLOCK);
    mutex_unlock(&dev->queue_lock);

    return ret;
}

static int vidioc_streamon(struct file *file, void *priv, enum v4l2_buf_type i)
{
    struct h8_video_dev *dev = video_drvdata(file);
    int ret;

    mutex_lock(&dev->queue_lock);
    dev->field_count = 0;
    ret = vb2_streamon(&dev->vb_vidq, i);
    mutex_unlock(&dev->queue_lock);

    return ret;
}

static int vidioc_streamoff(struct file *file, void *priv, enum v4l2_buf_type i)
{
    struct h8_video_dev *dev = video_drvdata(file);
    int ret;

    mutex_lock(&dev->queue_lock);
    ret = vb2_streamoff(&dev->vb_vidq, i);
    mutex_unlock(&dev->queue_lock);

    return ret;
}

static int vidioc_s_std(struct file *file, void *priv, v4l2_std_id std)
{
    //printk("kmd: vidioc_s_std\n");
    return 0;
}

static int vidioc_enum_input(struct file *file, void *priv, struct v4l2_input *inp)
{
    if (inp->index >= 2)
        return -EINVAL;

    //printk("kmd: vidioc_enum_input\n");

    inp->type   = V4L2_INPUT_TYPE_CAMERA;
    sprintf(inp->name, "Camera %u", inp->index);
    return 0;
}

static int vidioc_g_input(struct file *file, void *priv, unsigned int *i)
{
    struct h8_video_dev *dev = video_drvdata(file);

    //printk("kmd: vidioc_g_input\n");

    *i = dev->input;
    return 0;
}

static int vidioc_s_input(struct file *file, void *priv, unsigned int i)
{
    struct h8_video_dev *dev = video_drvdata(file);

    //printk("kmd: vidioc_s_input\n");

    if (i >= 2)
        return -EINVAL;

    dev->input = i;
    return 0;
}

#define  H8_SP_DATA_CTRL(CTL_DST, CTL_SIZE) \

#if 0
        if (0 == ctrl->val) \
            return 0; \
        if (copy_from_user(CTL_DST, (void*)ctrl->val, CTL_SIZE)) {\
            printk("V4L2: copy user space data fail. \n ");  return -EFAULT; \
        } \
        V4L2_H8_LOG(video_nr, h8_nr_VIDIOC_S_CTRL); \
        V4L2_H8_LOG_CTRL(video_nr, ctrl->name, ctrl->id - V4L2_CID_H8_CUSTOM_BASE); \
        V4L2_H8_LOG_DATA(video_nr, (unsigned int *)(CTL_DST), (CTL_SIZE)/4); \
        ctrl->cur.val = 0; ctrl->val = 0
#endif
// --- controls ----------------------------------------------
static int vip_s_ctrl(struct v4l2_ctrl *ctrl)
{
    struct h8_video_dev *dev = container_of(ctrl->handler, struct h8_video_dev, ctrl_handler);
    h8v_ctrl_t           func;
    int                  i;

    if (dev->skip_ctrl_hw)
    {
        //printk("V4l2: s_ctrl default\n");
        return 0;
    }

    switch (ctrl->id)
    {
        case V4L2_CID_STREAM_INIT:
        {
            if (!dev->stream_stoped)
                break;
            
            H8_SP_DATA_CTRL(&dev->stream_init_data, sizeof(Stream_Init));

            break;
        }
        case V4L2_CID_MAIN_CTRL:
        {
            // need sync with irq work.... or using a new copy. 
            // to do ...
            H8_SP_DATA_CTRL(&dev->main_ctrl_data, sizeof(ISP_Ctrl));

            break;
        }
        case V4L2_CID_TONEMAP_CURVE_SET:
        {
            H8_SP_DATA_CTRL(&dev->tonemap_curve_data, 295*4);
            break;
        }
        default:
        {
            printk("KMD: AP should not set this ctrl %s.\n", ctrl->name);
            return 0;
        }
    }

    func = h8v_ctrls[V4L2_H8_CTRL_NR(ctrl->id)];
    func(dev);

    return 0;
}

// -------------------------------------------------------------------
static const struct v4l2_ctrl_ops vip_ctrl_ops = {
    .s_ctrl     = vip_s_ctrl,
};

#define CTRL_NEW_CUSTOM(CTL_NAME) dev->CTL_NAME  = v4l2_ctrl_new_custom(hdl, &h8v_ctrl_##CTL_NAME, NULL)
#define MAKE_CTRL_CFG(CTL_NAME, CTL_ID, CTL_MIN, CTL_MAX) static const struct v4l2_ctrl_config h8v_ctrl_##CTL_NAME = { \
    .ops    = &vip_ctrl_ops,   \
    .id     = CTL_ID,  \
    .name   = #CTL_NAME,  \
    .type   = V4L2_CTRL_TYPE_INTEGER,   \
    .def    = 0x0,  \
    .min    = CTL_MIN,   \
    .max    = CTL_MAX,   \
    .step   = 1,    \
};


MAKE_CTRL_CFG(tonemap_curve,    V4L2_CID_TONEMAP_CURVE_SET,         0x80000000,                         0x7fffffff)


MAKE_CTRL_CFG(stream_init,      V4L2_CID_STREAM_INIT,               0x80000000,                         0x7fffffff)
MAKE_CTRL_CFG(main_ctrl,        V4L2_CID_MAIN_CTRL,                 0x80000000,                         0x7fffffff)
// -------------------------------------------------------------------
static const struct v4l2_file_operations vip_fops = {
    .owner          = THIS_MODULE,
    .open           = vip_video_open,
    .release        = vip_video_release,
    .poll           = vip_video_poll,
    .unlocked_ioctl = video_ioctl2,
    .mmap           = vip_file_mmap,
};

//UMD interface
static const struct v4l2_ioctl_ops vip_ioctl_ops = {
    .vidioc_querycap            = vidioc_querycap,
    .vidioc_enum_fmt_vid_cap    = vidioc_enum_fmt_vid_cap,
    .vidioc_g_fmt_vid_cap       = vidioc_g_fmt_vid_cap,
    .vidioc_try_fmt_vid_cap     = vidioc_try_fmt_vid_cap,
    .vidioc_s_fmt_vid_cap       = vidioc_s_fmt_vid_cap,

    .vidioc_reqbufs             = vidioc_reqbufs,
    .vidioc_querybuf            = vidioc_querybuf,
    .vidioc_qbuf                = vidioc_qbuf,
    .vidioc_dqbuf               = vidioc_dqbuf,

    .vidioc_s_std               = vidioc_s_std,
    .vidioc_enum_input          = vidioc_enum_input,
    .vidioc_g_input             = vidioc_g_input,
    .vidioc_s_input             = vidioc_s_input,
    
    .vidioc_streamon            = vidioc_streamon,
    .vidioc_streamoff           = vidioc_streamoff,

    .vidioc_enum_framesizes     = vidioc_enum_framesizes,
    .vidioc_enum_frameintervals = vidioc_enum_frameintervals,

    .vidioc_log_status          = v4l2_ctrl_log_status,
    .vidioc_subscribe_event     = v4l2_ctrl_subscribe_event,
    .vidioc_unsubscribe_event   = v4l2_event_unsubscribe,
};

static struct video_device vip_video_template = {
    .name           = "vip_v4l2",
    .fops           = &vip_fops,
    .ioctl_ops      = &vip_ioctl_ops,
    .minor          = -1,
    .release        = video_device_release,
    .tvnorms        = V4L2_STD_525_60,
};

static int vip_device_register(struct platform_device *pdev, int video_num)
{
    struct h8_video_dev       *dev;
    int                        index, ret;
    struct v4l2_ctrl_handler   *hdl = NULL;
    struct vb2_queue           *q   = NULL;

    dev = kzalloc(sizeof(*dev), GFP_KERNEL);
    if (!dev)
        return -ENOMEM;

    dev->id     = pdev->id;
    dev->pdev   = pdev;
    dev->video_index = video_num;
    g_vip_dev[video_num] = dev;

    //for pcie dma buffer
    if (!pdev->dev.dma_mask)
        pdev->dev.dma_mask = &pdev->dev.coherent_dma_mask;

    ret = dma_set_mask_and_coherent(&pdev->dev, DMA_BIT_MASK(32));
    if (ret)
        dprintk(1, "dma_set_mask_and_coherent fail\n");

    memcpy(dev->v4l2_dev.name, vendor_name, sizeof(*vendor_name));
    ret = v4l2_device_register(&pdev->dev, &(dev->v4l2_dev));
    if (ret < 0)
    {
        dprintk(1, "v4l2 device register fail, index %d\n", video_num);
        goto free_dev;
    }
    dev_set_drvdata(&(pdev)->dev, (dev));

    dev->fmt    = &formats[0];
    dev->width  = IMAGE_WIDTH;
    dev->height = IMAGE_HEIGHT;
    dev->field  = V4L2_FIELD_NONE;

    hdl = &dev->ctrl_handler;
    v4l2_ctrl_handler_init(hdl, (V4L2_CID_LAST_ITEM - V4L2_CID_H8_CUSTOM_BASE));    // num need be refined ....
    if (hdl->error)
    {
        dprintk(1, "hdl->error\n");
        ret = hdl->error;
        goto unreg_dev;
    }

    CTRL_NEW_CUSTOM(tonemap_curve);

    CTRL_NEW_CUSTOM(stream_init);
    CTRL_NEW_CUSTOM(main_ctrl);

    dev->v4l2_dev.ctrl_handler = hdl;

    /* initialize locks */
    spin_lock_init(&dev->slock);

    /* initialize queue */
    q           = &dev->vb_vidq;
    memset(q, 0, sizeof(dev->vb_vidq));
    q->type     = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    q->io_modes = VB2_DMABUF | VB2_USERPTR | VB2_MMAP | VB2_READ | VB2_WRITE;
    q->drv_priv = dev;
    q->buf_struct_size = sizeof(struct h8v_buffer);
    #if VIP_V4L2_VERSION_CODE >= KERNEL_VERSION(4,1,0)
    q->timestamp_flags = V4L2_BUF_FLAG_TIMESTAMP_MONOTONIC;
    #else
    q->timestamp_type = V4L2_BUF_FLAG_TIMESTAMP_MONOTONIC;
    #endif
    q->ops      = &vip_video_qops;
    q->mem_ops  = &vb2_vip_memops;//&vb2_dma_contig_memops; //&vb2_vmalloc_memops;
    //q->dev      = &pdev->dev;

    ret = vb2_queue_init(q);
    if (ret)
    {
        goto unreg_dev;
    }

	mutex_init(&dev->queue_lock);

    mutex_init(&dev->mutex);

    dev->video_dev = video_device_alloc();
    if (!dev->video_dev)
        goto unreg_dev;

    *(dev->video_dev) = vip_video_template;
    dev->video_dev->v4l2_dev = &dev->v4l2_dev;
    //set_bit(V4L2_FL_USE_FH_PRIO, &dev->video_dev->flags);
    #if VIP_V4L2_VERSION_CODE >= KERNEL_VERSION(5,4,0)
    dev->video_dev->device_caps = V4L2_CAP_VIDEO_CAPTURE | V4L2_CAP_STREAMING | V4L2_CAP_READWRITE;
    #endif
    dev_set_name(&dev->video_dev->dev, vendor_name);

    /*
	 * Provide a mutex to v4l2 core. It will be used to protect
	 * all fops and v4l2 ioctls.
	 */
	dev->video_dev->lock = &dev->mutex;

    video_set_drvdata(dev->video_dev, dev);

#if VIP_V4L2_VERSION_CODE < KERNEL_VERSION(5,7,0)
    ret = video_register_device(dev->video_dev, VFL_TYPE_GRABBER, video_num);
#else
    ret = video_register_device(dev->video_dev, VFL_TYPE_VIDEO, video_num);
#endif
    if (ret < 0)
        goto rel_vdev;

    dprintk(1, "vip v4l2 driver init OK, index %d\n", video_num);
    return 0;

rel_vdev:
    video_unregister_device(dev->video_dev);
unreg_dev:
    if (hdl)
    {
        v4l2_ctrl_handler_free(hdl);
    }
    v4l2_device_unregister(&dev->v4l2_dev);
free_dev:
    kfree(dev);
    return -ENOMEM;
}

static int vip_device_unregister(int video_num)
{
    struct h8_video_dev       *dev;

    dev = g_vip_dev[video_num];
    if (!dev)
        return -1;

    video_unregister_device(dev->video_dev);
    v4l2_device_unregister(&dev->v4l2_dev);
    v4l2_ctrl_handler_free(&dev->ctrl_handler);

    dprintk(1, "vip v4l2 video %d exit OK\n", video_num);
    kfree(dev);
    dev = NULL;
    return 0;
}
static int vip_video_probe(struct platform_device *pdev)
{
    int index, ret;
    int video_nr = 0;

    for (index = 0; index < VIP_CAMERA_MAX_NUMBER; index++) {
        ret = vip_device_register(pdev, video_nr);
        if (ret < 0)
            dprintk(1, "vip v4l2 video %d init failed\n", video_nr);
        video_nr++;
    }

    dprintk(1, "vip v4l2 driver init OK\n");
    return 0;
}

static void video_exit(void)
{
    int index;
    int video_nr = 0;
    for (index = 0; index < VIP_CAMERA_MAX_NUMBER; index++) {
        vip_device_unregister(video_nr);
        video_nr++;
    }
}

static int vip_video_remove(struct platform_device *pdev)
{
    //struct h8_video_dev *dev=(struct h8_video_dev *)dev_get_drvdata(&(pdev)->dev);
    return 0;
}

void vip_video_dev_release(struct device *dev)
{
    return;
}


static struct platform_driver vip_video_driver = {
    .probe              = vip_video_probe,
    .remove             = vip_video_remove,
    //.suspend            = h8_video_suspend,
    //.resume             = h8_video_resume,
    //.shutdown           = h8_video_shutdown,
    //.id_table           = h8_video_driver_ids,
    .driver = {
        .name           = vendor_name,
        .owner          = THIS_MODULE,
    }
};

static struct platform_device vip_video_device[] = {
    [0] = {
        .name                   = vendor_name,
        .id                     = 0,
        //.num_resources          = ARRAY_SIZE(csi0_resource),
        //.resource               = h8_video_resource,
        .dev.release            = vip_video_dev_release,
    }
};


static int __init vip_video_init(void)
{
    int ret = 0;

    dprintk(1, "init vip video module\n");

    ret = platform_driver_register(&vip_video_driver);
    if (ret)
    {
        dprintk(1, "platform driver register failed\n");
        return -1;
    }

    ret = platform_device_register(&vip_video_device[0]);
    if (ret)
    {
        dprintk(1, "platform device register failed\n");
        return -1;
    }

    return 0;
}

static void __exit vip_video_exit(void)
{
    //int ret;

    dprintk(1, "exit vip video module\n");

    video_exit();
    platform_device_unregister(&vip_video_device[0]);
    platform_driver_unregister(&vip_video_driver);

}

module_init(vip_video_init);
module_exit(vip_video_exit);

MODULE_LICENSE("Dual BSD/GPL");
MODULE_DESCRIPTION("V4l2 driver for VIP");
